class AutoRouter:
    # predefined method names and route rule expansions
    controller_action_map = {
        "list": ["/{slug}", ["GET"], "_list"],
        "show": ["/{slug}/<{type}:id_record>", ["GET"], "_show"],
        "create": ["/{slug}", ["POST"], "_create"],
        "update": ["/{slug}/<{type}:id_record>", ["PUT", "PATCH"], "_update"],
        "delete": ["/{slug}/<{type}:id_record>", ["DELETE"], "_delete"],
    }

    # predefined resource methods and route rule expansions
    resource_action_map = {
        # method_name: [rule, rule, ...]
        "get": [
            ["/{slug}", ["GET"], "_list"],
            ["/{slug}/<{type}:id_record>", ["GET"], "_show"],
        ],
        "post": [["/{slug}", ["POST"], "_create"]],
        "put": [["/{slug}/<{type}:id_record>", ["PUT", "PATCH"], "_update"]],
        "delete": [["/{slug}/<{type}:id_record>", ["DELETE"], "_delete"]],
    }

    @staticmethod
    def controller(app, slug: str, cls, id_type: str = "int"):
        """
        Register default routes for a controller class

        :param app: Flask app or blueprint
        :param slug: route slug
        :param cls: class to map
        :param id_type: optional datatype for id
        :return:
        """
        for method_name, opts in AutoRouter.controller_action_map.items():
            route, methods, suffix = opts
            if callable(getattr(cls, method_name, None)):
                app.add_url_rule(
                    route.format(slug=slug, type=id_type),
                    methods=methods,
                    view_func=cls.view_method(method_name),
                )

    @staticmethod
    def resource(app, slug, cls, id_type: str = None, prefix: str = ""):
        """
        Register default routes for a resource class

        Note: due to the simplistic nature of prefix hashing, if prefixes are used for multiple routes for the
        same class with different prefixes, there is a small chance of collision

        :param app: Flask app or blueprint
        :param slug: route slug
        :param cls: class to map
        :param id_type: optional datatype for id
        :return:
        """
        name = ".".join([cls.__module__, cls.__name__]).replace(".", "_")
        if not id_type:
            id_type = "int"
        for view_name, routes in AutoRouter.resource_action_map.items():
            for item in routes:
                route, methods, suffix = item
                if getattr(cls, view_name, None) is not None:
                    app.add_url_rule(
                        route.format(slug=slug, type=id_type),
                        methods=methods,
                        view_func=cls.as_view("{}{}".format(name, suffix)),
                    )



from rick.form import RequestRecord, field


class ShipperRequest(RequestRecord):
    fields = {
        "id": field(validators="required|id"),
        "name": field(validators="required|maxlen:40"),
        "phone": field(validators="maxlen:24"),
    }



import os
from argparse import ArgumentParser
from pathlib import Path
from pokie.codegen.template import TemplateProcessor
from pokie.core import CliCommand


class TplGenCommand(CliCommand):
    def arguments(self, parser: ArgumentParser):
        parser.add_argument("name", type=str, help="new module name")
        parser.add_argument("path", type=str, help="where to create the module")

    def get_template_path(self, args) -> Path:
        raise RuntimeError("abstract method")

    def get_vars(self, args):
        return {}

    def run(self, args) -> bool:
        dest_path = Path(args.path) / Path(args.name)
        if dest_path.exists():
            self.tty.error(
                "error: directory '{}' already exists".format(str(dest_path))
            )
            return False

        tpl_path = self.get_template_path(args)
        vars = self.get_vars(args)

        self.tty.write(self.tty.colorizer.white("generating structure..."))
        processor = TemplateProcessor([tpl_path])
        processor.process(tpl_path, dest_path, vars, self.tty)
        self.tty.write(self.tty.colorizer.green("template processed sucessfully!"))
        return True


class ModuleGenCmd(TplGenCommand):
    description = "create module structure"

    def get_template_path(self, args) -> Path:
        return (
            Path(os.path.dirname(__file__))
            / Path("..")
            / Path("template")
            / Path("module")
        )

    def get_vars(self, args):
        return {"{module_name}": args.name, "{ModuleName}": args.name.capitalize()}



import pytest

from pokie.constants import DEFAULT_LIST_SIZE
from pokie.http import DbGridRequest
from pokie_test.dto import CustomerRecord


@pytest.fixture
def dbgrid_request():
    return DbGridRequest(CustomerRecord)


@pytest.fixture
def dbgridRequest():
    # camel-case DbGridRequest()
    return DbGridRequest(CustomerRecord, use_camel_case=True)


class TestDbGrid:
    def test_validate(self, dbgrid_request):
        # empty request is valid
        assert dbgrid_request.is_valid({}) is True

        # default parameters with empty request
        data = dbgrid_request.dbgrid_parameters()
        assert data["search_text"] is None
        assert data["match_fields"] is None
        assert data["limit"] is None
        assert data["sort_fields"] is None
        assert data["search_fields"] is None

        # default parameters with limit
        data = dbgrid_request.dbgrid_parameters(DEFAULT_LIST_SIZE)
        assert data["limit"] == DEFAULT_LIST_SIZE

        # default parameters with search field names
        names = ["name1", "name2", "name3"]
        data = dbgrid_request.dbgrid_parameters(DEFAULT_LIST_SIZE, names)
        assert data["limit"] == DEFAULT_LIST_SIZE
        assert data["search_fields"] == names

    def test_validate_limit(self, dbgrid_request):
        # invalid limit
        assert dbgrid_request.is_valid({"limit": "abc"}) is False
        assert len(dbgrid_request.errors) == 1
        assert "limit" in dbgrid_request.errors.keys()

        assert dbgrid_request.is_valid({"limit": "-1"}) is False
        assert len(dbgrid_request.errors) == 1
        assert "limit" in dbgrid_request.errors.keys()

        assert dbgrid_request.is_valid({"limit": ""}) is False
        assert len(dbgrid_request.errors) == 1
        assert "limit" in dbgrid_request.errors.keys()

        # valid limits
        assert dbgrid_request.is_valid({"limit": None}) is True
        assert dbgrid_request.is_valid({"limit": "3"}) is True
        assert dbgrid_request.is_valid({"limit": 3}) is True

    def test_validate_offset(self, dbgrid_request):
        # invalid offset
        assert dbgrid_request.is_valid({"offset": "abc"}) is False
        assert len(dbgrid_request.errors) == 1
        assert "offset" in dbgrid_request.errors.keys()

        assert dbgrid_request.is_valid({"offset": "-1"}) is False
        assert len(dbgrid_request.errors) == 1
        assert "offset" in dbgrid_request.errors.keys()

        assert dbgrid_request.is_valid({"offset": ""}) is False
        assert len(dbgrid_request.errors) == 1
        assert "offset" in dbgrid_request.errors.keys()

        # valid offsets
        assert dbgrid_request.is_valid({"offset": None}) is True
        assert dbgrid_request.is_valid({"offset": "3"}) is True
        assert dbgrid_request.is_valid({"offset": 3}) is True

    def test_match_fields(self, dbgrid_request):
        assert dbgrid_request.is_valid({"match": "|"}) is False
        assert len(dbgrid_request.errors) == 1
        assert "match" in dbgrid_request.errors.keys()

        assert dbgrid_request.is_valid({"match": ""}) is True

        # match error - invalid clauses
        for clause in [
            "abc",
            "abc|",
            "abc:|",
            ":|",
            "|:",
            "|abc:",
            "|:def",
            "id:3|applied|abc",
        ]:
            assert dbgrid_request.is_valid({"match": clause}) is False
            assert len(dbgrid_request.errors) == 1
            assert "match" in dbgrid_request.errors.keys()

        # match error - invalid names
        for clause in ["abc:def", "123:4", "a:1|b:2|c:3"]:
            assert dbgrid_request.is_valid({"match": clause}) is False
            assert len(dbgrid_request.errors) == 1
            assert "match" in dbgrid_request.errors.keys()

        # match success - names
        # Note that no type validation is performed
        for clause in [
            "id:def",
            "id:1|contact_name:true",
            "id:1|contact_name:true|country:john",
        ]:
            assert dbgrid_request.is_valid({"match": clause}) is True

        # match fail - names are camelCased
        for clause in ["id:1|contactName:abc|contactTitle:xyz|postalCode:4400"]:
            assert dbgrid_request.is_valid({"match": clause}) is False
            # @todo: fix this

    def test_validate_match_fields_camelcase(self, dbgridRequest):
        # empty request is valid
        assert dbgridRequest.is_valid({}) is True

        # match success - names
        # Note that no type validation is performed
        for clause in [
            "id:def",
            "id:1|contactName:abc|contactTitle:xyz|postalCode:4400",
        ]:
            assert dbgridRequest.is_valid({"match": clause}) is True

    def test_validate_sort_fields(self, dbgrid_request):
        assert dbgrid_request.is_valid({"sort": "|"}) is False
        assert len(dbgrid_request.errors) == 1
        assert "sort" in dbgrid_request.errors.keys()

        assert dbgrid_request.is_valid({"osrt": ""}) is True

        # sort error - invalid clauses
        for clause in [
            "abc,",
            "abc:,",
            ":,",
            ",:",
            ",abc:",
            ",:def",
            "id:3,applied,abc",
        ]:
            assert dbgrid_request.is_valid({"sort": clause}) is False
            assert len(dbgrid_request.errors) == 1
            assert "sort" in dbgrid_request.errors.keys()

        # sort error - invalid names
        for clause in ["abc", "abc:asc" "123:4", "a:1,b:2,c:3"]:
            assert dbgrid_request.is_valid({"sort": clause}) is False
            assert len(dbgrid_request.errors) == 1
            assert "sort" in dbgrid_request.errors.keys()

        # sort success - names
        for clause in [
            "id",
            "id:asc,contact_name:desc",
            "id:asc,contact_name:desc,postal_code",
        ]:
            assert dbgrid_request.is_valid({"sort": clause}) is True

        # sort fail - camelCase names
        for clause in ["id:asc,contactName:desc", "id:asc,contactName:desc,postalCode"]:
            assert dbgrid_request.is_valid({"sort": clause}) is False
            assert len(dbgrid_request.errors) == 1
            assert "sort" in dbgrid_request.errors.keys()

    def test_validate_sort_fields_camelcase(self, dbgridRequest):
        # empty request is valid
        assert dbgridRequest.is_valid({}) is True

        # sort success
        for clause in ["contactName:asc,contactTitle:desc,postalCode", "contactName"]:
            assert dbgridRequest.is_valid({"sort": clause}) is True



from typing import Optional

from rick.base import Di
from rick.mixin import Injectable

from pokie.constants import DI_DB
from pokie_test.dto import CustomerRecord
from pokie_test.repository import CustomerRepository


class CustomerService(Injectable):
    def __init__(self, di: Di):
        super().__init__(di)

    def get_customer(self, id_customer: str) -> Optional[CustomerRecord]:
        return self.repo_customer().fetch_pk(id_customer)

    def repo_customer(self) -> CustomerRepository:
        return CustomerRepository(self.get_di().get(DI_DB))



from typing import Optional, Any
from flask.typing import ResponseReturnValue

from pokie.http import PokieView


class HookView(PokieView):
    MSG = "the quick brown fox jumps over the lazy dog"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # attribute to be filled by the hook
        self.sample = None
        # add hook to the hook list
        self.dispatch_hooks.append("_hook_example")

    def _hook_example(
        self, method: str, *args: Any, **kwargs: Any
    ) -> Optional[ResponseReturnValue]:
        self.sample = HookView.MSG
        # hook was successful, returns None
        return None
        # if hook was unsuccessful, it must return a Response
        # return self.forbidden()

    def get(self):
        # return value injected by hook
        return self.success(self.sample)



import inspect
import json
from argparse import ArgumentParser

from tabulate import tabulate

from pokie.constants import DI_APP
from pokie.core import CliCommand


class ModuleListCmd(CliCommand):
    description = "list loaded modules"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument(
            "--json",
            help="output JSON instead of tabular format",
            action="store_true",
            default=False,
        )

    def run(self, args) -> bool:
        table = []
        for module_name, module in self.get_di().get(DI_APP).modules.items():
            table.append(
                [
                    module_name,
                    inspect.getfile(module.__class__),
                    module.territory_description,
                ]
            )

        if not args.json:
            self.tty.write(
                tabulate(table, headers=["Name", "Module class path", "Description"])
            )
        else:
            result = []
            for row in table:
                result.append(
                    {"name": row[0], "moduleClassPath": row[1], "description": row[2]}
                )
            self.tty.write(json.dumps(result, indent=2))
        return True



from .controller_view import CustomerController
from .custom_requestrecord import CustomRequestRecordView
from .custom_response import CustomResponseView
from .dispatch_hook import HookView
from .camel_case import CamelCaseResponseView



from typing import Any, Optional

from flask.typing import ResponseReturnValue
from pokie.http import PokieView


class MyInitView(PokieView):
    init_methods = ["foo"]

    def foo(self, **kwargs):
        self.test = [1]
        # add custom hook
        self.internal_hooks.append("myhook")

    def myhook(
        self, method: str, *args: Any, **kwargs: Any
    ) -> Optional[ResponseReturnValue]:
        self.test.append(2)
        return None

    def get(self):
        pass


class TestViewAdvanced:
    def test_view_init(self, pokie_app):
        with pokie_app.app_context():
            view = MyInitView()
            assert getattr(view, "test", None) is not None
            assert getattr(view, "test", None) == [1]

    def test_view_dispatch(self, pokie_app):
        with pokie_app.app_context():
            with pokie_app.test_request_context():
                view = MyInitView()
                view.dispatch_request()
                assert getattr(view, "test", None) == [1, 2]



from rick.form import RequestRecord, field


class StatesRequest(RequestRecord):
    fields = {
        "id": field(validators="required|id"),
        "name": field(validators="required|maxlen:100"),
        "abbr": field(validators="required|maxlen:2"),
        "region": field(validators="required|maxlen:50"),
    }



from rick.form import RequestRecord, field

from pokie.rest import RestView
from pokie_test.dto import CustomerRecord


# RequestRecord class
class CustomerRequest(RequestRecord):
    fields = {
        "id": field(validators="required|maxlen:16"),
        "company_name": field(validators="required|maxlen:40"),
        "contact_name": field(validators="maxlen:30"),
        "contact_title": field(validators="maxlen:30"),
        "address": field(validators="maxlen:60"),
        "city": field(validators="maxlen:15"),
        "region": field(validators="maxlen:15"),
        "postal_code": field(validators="maxlen:15"),
        "country": field(validators="maxlen:15"),
        "phone": field(validators="maxlen:24"),
        "fax": field(validators="maxlen:24"),
    }


class CustomerView(RestView):
    # RequestRecord class for body operations
    request_class = CustomerRequest

    # Database Record class
    record_class = CustomerRecord

    # allowed search fields to be used with the search variable
    search_fields = [
        CustomerRecord.contact_name,
        CustomerRecord.company_name,
    ]

    # optional custom service name to be used; if no service name is specified, an instance of
    # pokie.rest.RestService is automatically created
    # service_name = 'my-service-name'

    # optional limit for default listing operations
    # if list_limit > 0, the specified value will be used as default limit for unbounded listing requests
    # list_limit = -1



from argparse import ArgumentParser

from pokie.constants import DI_SERVICES, DI_APP
from pokie.contrib.base.constants import SVC_FIXTURE
from pokie.contrib.base.dto import FixtureRecord
from pokie.contrib.base.service.fixture import FixtureService
from pokie.core import CliCommand


class FixtureCmd(CliCommand):
    @property
    def svc_fixture(self) -> FixtureService:
        return self.get_di().get(DI_SERVICES).get(SVC_FIXTURE)


class RunFixtureCmd(FixtureCmd):
    description = "run existing fixtures"

    def arguments(self, parser: ArgumentParser):
        parser.add_argument("name", type=str, help="fixture name(s) to run", nargs="*")

    def valid_name(self, name: str) -> bool:
        return name.find(".") > -1

    def run(self, args) -> bool:
        existing = []
        all = self.svc_fixture.scan() if len(args.name) == 0 else args.name

        for r in self.svc_fixture.list():
            existing.append(r.name)

        for name in all:
            if self.valid_name(name):
                self.tty.write("Fixture {}: ".format(name), eol=False)
                if name in existing:
                    self.tty.write(
                        self.tty.colorizer.white(
                            "already executed, skipping", attr="bold"
                        )
                    )
                else:
                    try:
                        self.svc_fixture.execute(name)
                        self.svc_fixture.add(FixtureRecord(name=name))
                        self.tty.write(
                            self.tty.colorizer.green(
                                "executed successfully", attr="bold"
                            )
                        )
                    except Exception as e:
                        self.tty.error("\nError : " + str(e))
                        return False
            else:
                self.tty.error("Fixture '{}': invalid name, skipping".format(name))
        return True


class CheckFixtureCmd(FixtureCmd):
    description = "show existing fixture status"

    def run(self, args) -> bool:
        existing = []
        all = self.svc_fixture.scan()
        if len(all) != len(set(all)):
            # @todo: use rick.util.misc.list_duplicates()
            duplicates = []
            seen = set()
            for item in all:
                if item in seen:
                    duplicates.append(item)
                else:
                    seen.add(item)
            self.tty.error(
                "Duplicated fixture(s) found: {}".format(",".join(duplicates))
            )
            return False

        for r in self.svc_fixture.list():
            existing.append(r.name)

        for name in all:
            self.tty.write("Fixture {}: ".format(name), eol=False)
            if name in existing:
                self.tty.write(
                    self.tty.colorizer.white("already executed, skipping", attr="bold")
                )
            else:
                self.tty.write(self.tty.colorizer.green("new fixture", attr="bold"))
        return True



import json
from flask import Response


class PokieResponse:
    def __init__(self, response: Response):
        self.success = True
        self.error = None
        self.data = None
        self.error_message = ""
        self.form_error = {}
        self.code = 0

        if response is None:
            return

        self.code = response.status_code
        self.success = response.status_code == 200

        if response.data:
            data = json.loads(response.data)

            if "data" in data.keys():
                self.data = data["data"]

            if "error" in data.keys():
                self.error = data["error"]
                if "message" in self.error.keys():
                    self.error_message = self.error["message"]
                if "formError" in self.error.keys():
                    self.form_error = self.error["formError"]


class AuthInterface:
    def authenticate(self, client) -> bool:
        raise NotImplementedError


class PokieClient:
    def __init__(self, client, base_url="", auth: AuthInterface = None):
        self.client = client
        self.base_url = base_url
        self.headers = {"Content-type": "application/json"}
        self.auth = None
        if auth is not None:
            self.auth = auth
            self.auth.authenticate(self)

    def post(self, url: str, data=None) -> PokieResponse:
        return PokieResponse(
            self.client.post(
                "{}{}".format(self.base_url, url), json=data, headers=self.headers
            )
        )

    def patch(self, url: str, data=None) -> PokieResponse:
        return PokieResponse(
            self.client.patch(
                "{}{}".format(self.base_url, url), json=data, headers=self.headers
            )
        )

    def raw_get(self, url: str, data=None) -> Response:
        return self.client.get(
            "{}{}".format(self.base_url, url), json=data, headers=self.headers
        )

    def post_formdata(self, url: str, data=None) -> PokieResponse:
        return PokieResponse(
            self.client.post(
                "{}{}".format(self.base_url, url),
                data=data,
                content_type="multipart/form-data",
            )
        )

    def get(self, url: str, data=None) -> PokieResponse:
        return PokieResponse(
            self.client.get(
                "{}{}".format(self.base_url, url), json=data, headers=self.headers
            )
        )

    def put(self, url: str, data=None) -> PokieResponse:
        return PokieResponse(
            self.client.put(
                "{}{}".format(self.base_url, url), json=data, headers=self.headers
            )
        )

    def delete(self, url: str, data=None) -> PokieResponse:
        return PokieResponse(
            self.client.delete(
                "{}{}".format(self.base_url, url), json=data, headers=self.headers
            )
        )



from rick.form import RequestRecord, field

from pokie.codegen.pg import PgTableSpec
from pokie.codegen import RequestGenerator

tablespec_serial_record = """
from rick.form import RequestRecord, field

class TablespecSerialRequest(RequestRecord):
    fields = {
        'id': field(validators='id|numeric', bind='id'),
        'field_bigint': field(validators='numeric', bind='field_bigint'),
        'field_bigserial': field(validators='required|numeric', bind='field_bigserial'),
        'field_bit': field(validators='', bind='field_bit'),
        'field_varbit': field(validators='', bind='field_varbit'),
        'field_box': field(validators='', bind='field_box'),
        'field_bytea': field(validators='', bind='field_bytea'),
        'field_char': field(validators='maxlen:10', bind='field_char'),
        'field_varchar': field(validators='maxlen:10', bind='field_varchar'),
        'field_cidr': field(validators='', bind='field_cidr'),
        'field_circle': field(validators='', bind='field_circle'),
        'field_date': field(validators='iso8601', bind='field_date'),
        'field_float8': field(validators='', bind='field_float8'),
        'field_inet': field(validators='', bind='field_inet'),
        'field_int4': field(validators='numeric', bind='field_int4'),
        'field_interval': field(validators='', bind='field_interval'),
        'field_json': field(validators='', bind='field_json'),
        'field_jsonb': field(validators='', bind='field_jsonb'),
        'field_line': field(validators='', bind='field_line'),
        'field_lseg': field(validators='', bind='field_lseg'),
        'field_macaddr': field(validators='', bind='field_macaddr'),
        'field_macaddr8': field(validators='', bind='field_macaddr8'),
        'field_money': field(validators='', bind='field_money'),
        'field_numeric': field(validators='decimal', bind='field_numeric'),
        'field_path': field(validators='', bind='field_path'),
        'field_point': field(validators='', bind='field_point'),
        'field_polygon': field(validators='', bind='field_polygon'),
        'field_float4': field(validators='', bind='field_float4'),
        'field_int2': field(validators='numeric', bind='field_int2'),
        'field_smallserial': field(validators='required|numeric', bind='field_smallserial'),
        'field_text': field(validators='', bind='field_text'),
        'field_time': field(validators='', bind='field_time'),
        'field_timetz': field(validators='', bind='field_timetz'),
        'field_timestamp': field(validators='iso8601', bind='field_timestamp'),
        'field_timestamptz': field(validators='iso8601', bind='field_timestamptz'),
        'field_tsquery': field(validators='', bind='field_tsquery'),
        'field_tsvector': field(validators='', bind='field_tsvector'),
        'field_uuid': field(validators='', bind='field_uuid'),
        'field_xml': field(validators='', bind='field_xml'),   
    }
"""

tablespec_serial_record_camel = """
from rick.form import RequestRecord, field

class TablespecSerialRequest(RequestRecord):
    fields = {
        'id': field(validators='id|numeric', bind='id'),
        'fieldBigint': field(validators='numeric', bind='field_bigint'),
        'fieldBigserial': field(validators='required|numeric', bind='field_bigserial'),
        'fieldBit': field(validators='', bind='field_bit'),
        'fieldVarbit': field(validators='', bind='field_varbit'),
        'fieldBox': field(validators='', bind='field_box'),
        'fieldBytea': field(validators='', bind='field_bytea'),
        'fieldChar': field(validators='maxlen:10', bind='field_char'),
        'fieldVarchar': field(validators='maxlen:10', bind='field_varchar'),
        'fieldCidr': field(validators='', bind='field_cidr'),
        'fieldCircle': field(validators='', bind='field_circle'),
        'fieldDate': field(validators='iso8601', bind='field_date'),
        'fieldFloat8': field(validators='', bind='field_float8'),
        'fieldInet': field(validators='', bind='field_inet'),
        'fieldInt4': field(validators='numeric', bind='field_int4'),
        'fieldInterval': field(validators='', bind='field_interval'),
        'fieldJson': field(validators='', bind='field_json'),
        'fieldJsonb': field(validators='', bind='field_jsonb'),
        'fieldLine': field(validators='', bind='field_line'),
        'fieldLseg': field(validators='', bind='field_lseg'),
        'fieldMacaddr': field(validators='', bind='field_macaddr'),
        'fieldMacaddr8': field(validators='', bind='field_macaddr8'),
        'fieldMoney': field(validators='', bind='field_money'),
        'fieldNumeric': field(validators='decimal', bind='field_numeric'),
        'fieldPath': field(validators='', bind='field_path'),
        'fieldPoint': field(validators='', bind='field_point'),
        'fieldPolygon': field(validators='', bind='field_polygon'),
        'fieldFloat4': field(validators='', bind='field_float4'),
        'fieldInt2': field(validators='numeric', bind='field_int2'),
        'fieldSmallserial': field(validators='required|numeric', bind='field_smallserial'),
        'fieldText': field(validators='', bind='field_text'),
        'fieldTime': field(validators='', bind='field_time'),
        'fieldTimetz': field(validators='', bind='field_timetz'),
        'fieldTimestamp': field(validators='iso8601', bind='field_timestamp'),
        'fieldTimestamptz': field(validators='iso8601', bind='field_timestamptz'),
        'fieldTsquery': field(validators='', bind='field_tsquery'),
        'fieldTsvector': field(validators='', bind='field_tsvector'),
        'fieldUuid': field(validators='', bind='field_uuid'),
        'fieldXml': field(validators='', bind='field_xml'),
    }
"""

tablespec_rel_record = """
from rick.form import RequestRecord, field

class TablespecRelRequest(RequestRecord):
    fields = {
        'id': field(validators='id|numeric', bind='id'),
        'fk_tablespec_serial': field(validators='required|numeric|pk:public.tablespec_serial,id_tablespec', bind='fk_tablespec_serial'),
    }
"""

tablespec_rel_record_camel = """
from rick.form import RequestRecord, field

class TablespecRelRequest(RequestRecord):
    fields = {
        'id': field(validators='id|numeric', bind='id'),
        'fkTablespecSerial': field(validators='required|numeric|pk:public.tablespec_serial,id_tablespec', bind='fk_tablespec_serial'),
    }
"""
tablespec_serial_names = [
    "id",
    "field_bigint",
    "field_bigserial",
    "field_bit",
    "field_varbit",
    "field_box",
    "field_bytea",
    "field_char",
    "field_varchar",
    "field_cidr",
    "field_circle",
    "field_date",
    "field_float8",
    "field_inet",
    "field_int4",
    "field_interval",
    "field_json",
    "field_jsonb",
    "field_line",
    "field_lseg",
    "field_macaddr",
    "field_macaddr8",
    "field_money",
    "field_numeric",
    "field_path",
    "field_point",
    "field_polygon",
    "field_float4",
    "field_int2",
    "field_smallserial",
    "field_text",
    "field_time",
    "field_timetz",
    "field_timestamp",
    "field_timestamptz",
    "field_tsquery",
    "field_tsvector",
    "field_uuid",
    "field_xml",
]


class TestRequestGenerator:
    def test_generate_source(self, pokie_db):
        pg_spec = PgTableSpec(pokie_db)
        spec = pg_spec.generate("tablespec_serial")
        generator = RequestGenerator()
        # default settings
        src = generator.generate_source(spec)
        assert self._cleanup(src) == self._cleanup(tablespec_serial_record)

        # camelCase
        src = generator.generate_source(spec, camelcase=True)
        assert self._cleanup(src) == self._cleanup(tablespec_serial_record_camel)

        # default settings -rel
        spec = pg_spec.generate("tablespec_rel")
        src = generator.generate_source(spec)
        assert self._cleanup(src) == self._cleanup(tablespec_rel_record)

        # camelCase
        src = generator.generate_source(spec, camelcase=True)
        assert self._cleanup(src) == self._cleanup(tablespec_rel_record_camel)

    def test_generate_class(self, pokie_db):
        pg_spec = PgTableSpec(pokie_db)
        spec = pg_spec.generate("tablespec_serial")
        generator = RequestGenerator()

        cls = generator.generate_class(spec)
        assert issubclass(cls, RequestRecord)
        for name in tablespec_serial_names:
            assert name in cls.fields.keys()

    def _cleanup(self, s: str):
        for c in ["\n", " "]:
            s = s.replace(c, "")
        return c



from threading import Lock


class ModuleRunnerMiddleware:
    def __init__(self, app, pokie_app):
        self.app = app
        self.pokie_app = pokie_app
        self.lock = Lock()

    def __call__(self, environ, start_response):
        self.pokie_app.init()
        return self.app(environ, start_response)



from pokie.core import BaseModule
from pokie.http import AutoRouter, PokieView
from pokie.rest.auto import Auto
from pokie_test.constants import SVC_NORTHWIND_CUSTOMER, SVC_NORTHWIND_STATES
from pokie_test.dto import (
    CategoryRecord,
    ShipperRecord,
    UsStatesRecord,
    TerritoriesRecord,
)
from pokie_test.views import (
    CustomerController,
    CustomRequestRecordView,
    CustomResponseView,
    CamelCaseResponseView,
)
from pokie_test.views.dispatch_hook import HookView
from pokie_test.views.northwind_customer import CustomerView

from pokie_test.views.northwind_shipper import ShipperRequest
from pokie_test.views.northwind_states import StatesRequest
from pokie_test.views.northwind_territories import TerritoriesView


class Module(BaseModule):
    # internal module name
    name = "pokie_test"

    # internal module description
    description = "mock module for pokie unit testing"

    # service map
    #
    # this map registers existing module services, and makes them available to the application. Services are lazy-loaded
    # upon first invocation, to reduce overhead. A service class must inherit the Injectable mixin.
    #
    # The service mapper will load services by service name; As such, service names should be unique, unless the goal
    # is to explicitly override already-registered services
    #
    services = {
        # service entries are defined as {'service_name': 'path_to_class'}
        SVC_NORTHWIND_CUSTOMER: "pokie_test.service.CustomerService",
        SVC_NORTHWIND_STATES: "pokie_test.service.StatesService",
    }

    # cli command map
    #
    # this map registers existing cli commands exposed by the module. Cli commands are classes that inherit from
    # CliCommand. Cli command names should be unique, unless the goal is to explicitly override existing commands
    #
    cmd = {
        # cli commands are defined as {'command': 'path_to_class'}
        # 'cli_cmd': 'path.to.class'
    }

    # events map
    #
    # Pokie has a concept similar to signals in other framework., but with different capabilities, called Events. Events
    # are classes that extend from EventHandler, and implement a method with the name of the event. Event handler objects
    # are short-lived - they are created upon dispatched of a given event, and de-referenced afterwards.
    #
    # Events have a priority number - handler execution follows the priority number in descending order (lower numbers get
    # executed first)
    #
    # Events also have optional in and out objects, typically used for dictionary composition. A common use case is to add
    # extra information to the response generated on a given information, such as login
    #
    # Event names are unique strings that identify the event; there is no specific requirements for naming, but common
    # convention suggests the usage of snakecase (eg. some_event).
    #
    #
    # events are refined as a two-level structure, containing zero or more events, and then handlers and priorities:
    # events = {
    #   'event_name': {
    #       numeric_priority: [path_to_handler, path_to_handler, ...]
    #   }
    # }
    #
    events = {}

    # worker jobs list
    #
    # jobs are tasks that are executed continuously and cooperatively, in a closed loop; their purpose is to execute
    # background operations such as sending emails or resizing images. What makes them different from cron approaches
    # is their continuous execution - there is an idle job with a default 15s pause, and then all other jobs are run
    # sequentially, in a closed loop. If a given job takes too much time to execute, it will delay subsequent jobs, so
    # this approach may not fit all workloads.
    #
    # Jobs are long-lived objects whose class must extend Injectable and Runnable mixins.
    # The job list is a list of strings with the full path for each job class, similar to other existing referencing structures
    #
    jobs = [
        # 'full.path.to.job.class'
    ]

    fixtures = [
        "pokie_test.fixtures.ExampleFixture",
    ]

    def build(self, parent=None):
        # This method is called when modules are initialized; At this point, all other dependencies have already been
        # initialized, including the Service Manager, Event Manager and even registered factories
        #
        # All Flask-related routing calls should reside here
        app = parent.app
        AutoRouter.resource(app, "customers", CustomerView, id_type="string")

        app.add_url_rule(
            "/mycustomer/<string:id_customer>",
            methods=["GET"],
            view_func=CustomerController.view_method("view_customer"),
        )

        app.add_url_rule(
            "/views/custom-requestrecord",
            methods=["GET", "POST", "PUT", "PATCH"],
            view_func=CustomRequestRecordView.as_view("view_customrequestrecord"),
        )

        app.add_url_rule(
            "/views/custom-response",
            methods=["GET"],
            view_func=CustomResponseView.as_view("view_customresponse"),
        )

        app.add_url_rule(
            "/views/hooks",
            methods=["GET"],
            view_func=HookView.as_view("view_hooks", di=app.di),
        )

        app.add_url_rule(
            "/views/camelcase",
            methods=["GET"],
            view_func=CamelCaseResponseView.as_view("view_camelcase"),
        )

        # Auto Rest - category
        Auto.rest(
            app, "catalog/category", CategoryRecord, search_fields=[CategoryRecord.name]
        )

        # Auto Rest with custom RequestRecord class
        Auto.rest(
            app,
            "catalog/shipper",
            ShipperRecord,
            ShipperRequest,
            search_fields=[ShipperRecord.name],
        )

        # Auto Rest with custom RequestRecord and custom service
        Auto.rest(
            app,
            "region/states",
            UsStatesRecord,
            StatesRequest,
            SVC_NORTHWIND_STATES,
            search_fields=[
                UsStatesRecord.name,
                UsStatesRecord.abbr,
                UsStatesRecord.region,
            ],
        )

        # Auto Rest with custom base Class
        Auto.rest(
            app,
            "region/territories",
            TerritoriesRecord,
            base_cls=TerritoriesView,
            id_type="string",
            search_fields=[
                TerritoriesRecord.id,
                TerritoriesRecord.territory_description,
            ],
        )

        # Auto View - suppliers
        view = Auto.view(app, "suppliers")
        AutoRouter.resource(app, "catalog/supplier", view)

        # Auto View - products
        view = Auto.view(app, "products")
        AutoRouter.resource(app, "catalog/product", view)

        # Auto View - tablespec
        view = Auto.view(app, "tablespec_serial")
        AutoRouter.resource(app, "tablespec", view)



from pokie.contrib.base.constants import SVC_VALIDATOR, SVC_SETTINGS, SVC_FIXTURE
from pokie.contrib.base.validators import init_validators
from pokie.core import BaseModule


class Module(BaseModule):
    name = "base"
    description = "Pokie base module"

    cmd = {
        # base commands
        "list": "pokie.contrib.base.cli.ListCmd",
        "help": "pokie.contrib.base.cli.HelpCmd",
        "version": "pokie.contrib.base.cli.VersionCmd",
        "runserver": "pokie.contrib.base.cli.RunServerCmd",
        "module:list": "pokie.contrib.base.cli.ModuleListCmd",
        "route:list": "pokie.contrib.base.cli.RouteListCmd",
        # database-related commands
        "db:init": "pokie.contrib.base.cli.DbInitCmd",
        "db:check": "pokie.contrib.base.cli.DbCheckCmd",
        "db:update": "pokie.contrib.base.cli.DbUpdateCmd",
        # worker job commands
        "job:list": "pokie.contrib.base.cli.JobListCmd",
        "job:run": "pokie.contrib.base.cli.JobRunCmd",
        # code generation
        "codegen:dto": "pokie.contrib.base.cli.GenDtoCmd",
        "codegen:request": "pokie.contrib.base.cli.GenRequestRecordCmd",
        "codegen:module": "pokie.contrib.base.cli.ModuleGenCmd",
        # fixtures
        "fixture:run": "pokie.contrib.base.cli.RunFixtureCmd",
        "fixture:check": "pokie.contrib.base.cli.CheckFixtureCmd",
        # tests
        "pytest": "pokie.contrib.base.cli.PyTestCmd",
    }

    services = {
        # db-related validators
        SVC_VALIDATOR: "pokie.contrib.base.service.ValidatorService",
        # settings service
        SVC_SETTINGS: "pokie.contrib.base.service.SettingsService",
        # fixture service
        SVC_FIXTURE: "pokie.contrib.base.service.FixtureService",
    }

    jobs = [
        "pokie.contrib.base.job.IdleJob",
    ]

    fixtures = []

    def build(self, parent=None):
        init_validators(self.get_di())



from pokie.http import AutoRouter, PokieView
from pokie.rest import RestView


class IncompleteResourceView(PokieView):
    def get(self, id_record=None):
        pass


class CompleteResourceView(RestView):
    pass


class IncompleteControllerView(PokieView):
    def show(self, id_record):
        pass

    def create(self):
        pass


class CompleteControllerView(IncompleteControllerView):
    def list(self):
        pass

    def update(self, id_record):
        pass

    def delete(self, id_record):
        pass


class TestAutoRouter:
    def test_incomplete_resource(self, pokie_app):
        slug = "test_operation"
        # register incomplete resource class
        AutoRouter.resource(pokie_app, slug, IncompleteResourceView)

        result = {}
        with pokie_app.app_context():
            for rule in pokie_app.url_map.iter_rules():
                methods = rule.methods
                rule_name = str(rule)
                if rule_name.startswith("/" + slug):
                    result[rule_name] = sorted(methods)

        # should only have 2 endpoints - get and list
        assert len(result) == 2
        key = "/" + slug
        assert key in result.keys()
        assert result[key] == sorted({"HEAD", "GET", "OPTIONS"})
        key = "/" + slug + "/<int:id_record>"
        assert key in result.keys()
        assert result[key] == sorted({"HEAD", "GET", "OPTIONS"})

    def test_complete_resource(self, pokie_app):
        slug = "test_operation"
        # register complete
        AutoRouter.resource(pokie_app, slug, CompleteResourceView)

        result = []
        with pokie_app.app_context():
            for rule in pokie_app.url_map.iter_rules():
                methods = rule.methods
                rule_name = str(rule)
                if rule_name.startswith("/" + slug):
                    result.append([rule_name, sorted(methods)])

        expected = [
            ["/test_operation", ["GET", "HEAD", "OPTIONS"]],
            ["/test_operation/<int:id_record>", ["GET", "HEAD", "OPTIONS"]],
            ["/test_operation", ["OPTIONS", "POST"]],
            ["/test_operation/<int:id_record>", ["OPTIONS", "PATCH", "PUT"]],
            ["/test_operation/<int:id_record>", ["DELETE", "OPTIONS"]],
        ]

        # should only have 2 endpoints - get and list
        assert len(result) == len(expected)
        for rule in expected:
            assert rule in result

    def test_incomplete_controller(self, pokie_app):
        slug = "test_operation"
        # register incomplete controller class
        AutoRouter.controller(pokie_app, slug, IncompleteControllerView, "string")

        result = {}
        with pokie_app.app_context():
            for rule in pokie_app.url_map.iter_rules():
                methods = rule.methods
                rule_name = str(rule)
                if rule_name.startswith("/" + slug):
                    result[rule_name] = sorted(methods)

        # should only have 2 endpoints - get and list
        assert len(result) == 2
        key = "/" + slug
        assert key in result.keys()
        assert result[key] == sorted({"OPTIONS", "POST"})
        key = "/" + slug + "/<string:id_record>"
        assert key in result.keys()
        assert result[key] == sorted({"HEAD", "GET", "OPTIONS"})

    def test_complete_controller(self, pokie_app):
        slug = "test_operation"
        # register complete controller
        AutoRouter.controller(pokie_app, slug, CompleteControllerView)

        result = []
        with pokie_app.app_context():
            for rule in pokie_app.url_map.iter_rules():
                methods = rule.methods
                rule_name = str(rule)
                if rule_name.startswith("/" + slug):
                    result.append([rule_name, sorted(methods)])

        expected = [
            ["/test_operation", ["GET", "HEAD", "OPTIONS"]],
            ["/test_operation/<int:id_record>", ["GET", "HEAD", "OPTIONS"]],
            ["/test_operation", ["OPTIONS", "POST"]],
            ["/test_operation/<int:id_record>", ["OPTIONS", "PATCH", "PUT"]],
            ["/test_operation/<int:id_record>", ["DELETE", "OPTIONS"]],
        ]
        # should only have 2 endpoints - get and list
        assert len(result) == len(expected)
        for rule in expected:
            assert rule in result



from flask import Flask
from rick.base import Container, Di, MapLoader
from rick.base.container import ContainerBase
from rick.event import EventManager
from rick.resource.console import ConsoleWriter

from pokie.constants import (
    DI_CONFIG,
    DI_APP,
    DI_FLASK,
    DI_SIGNAL,
    DI_TTY,
    DI_SERVICES,
    DI_EVENTS,
    DI_HTTP_ERROR_HANDLER,
)
from pokie.core import FlaskApplication, SignalManager, BaseModule
from pokie.http import HttpErrorHandler
from pokie_test.constants import SVC_NORTHWIND_CUSTOMER


class TestFlaskApplication:
    def test_init(self):
        cfg = Container({})
        app = FlaskApplication(cfg)
        self.assert_app(app)

    def test_build(self):
        cfg = Container({})
        app = FlaskApplication(cfg)
        self.assert_app(app)

        app.build([], [])

        assert app.app is not None
        assert isinstance(app.app, Flask)
        assert app.app.di is not None
        assert isinstance(app.app.di, Di)
        assert isinstance(app.di.get(DI_FLASK), Flask)

        assert isinstance(app.di.get(DI_SIGNAL), SignalManager)
        assert isinstance(app.di.get(DI_TTY), ConsoleWriter)
        assert isinstance(app.di.get(DI_SERVICES), MapLoader)
        assert isinstance(app.di.get(DI_EVENTS), EventManager)
        assert app.di.has(DI_HTTP_ERROR_HANDLER) is False

    def test_current_app(self, pokie_app):
        app = pokie_app.di.get(DI_APP)  # type: FlaskApplication
        self.assert_app(app)

        # modules
        assert app.modules is not None
        assert len(app.modules) == 2
        assert isinstance(app.modules, dict)
        for name, obj in app.modules.items():
            assert name in ["pokie.contrib.base", "pokie_test"]
            assert isinstance(obj, BaseModule)

        # services
        svc = app.di.get(DI_SERVICES)
        assert svc is not None
        assert svc.contains(SVC_NORTHWIND_CUSTOMER)

        # http error handler
        assert app.di.has(DI_HTTP_ERROR_HANDLER)
        assert isinstance(app.di.get(DI_HTTP_ERROR_HANDLER), HttpErrorHandler)

    def assert_app(self, app: FlaskApplication):
        assert app is not None

        assert app.di is not None
        assert isinstance(app.di, Di)
        assert app.di.get(DI_CONFIG) is not None
        assert isinstance(app.di.get(DI_CONFIG), ContainerBase)
        assert app.di.get(DI_APP) is not None

        assert app.modules is not None
        assert isinstance(app.modules, dict)



from pokie.http import PokieView
from pokie_test.dto import CustomerRecord


class CamelCaseResponseView(PokieView):
    camel_case = True

    def get(self):
        record = CustomerRecord(
            company_name="company_name",
            contact_name="contact_name",
            contact_title="contact_title",
            address="address",
        )
        return self.success(record)



from pokie.constants import HTTP_OK, HTTP_BADREQ, HTTP_NOT_FOUND, DI_DB
from pokie.rest import RestService
from pokie.test import PokieClient
from pokie_test.dto import CustomerRecord
from pokie_test.repository import CustomerRepository


class TestAuto:
    def test_tablespec(self, pokie_app):
        with pokie_app.test_client() as client:
            client = PokieClient(client)

            # insert item
            record = {
                "field_bigint": 12,
                "field_bit": "1",
                "field_varbit": "10101",
                "field_bytea": "string",
                "field_char": "b",
            }
            result = client.post("/tablespec", data=record)
            assert result.code == HTTP_OK
            assert result.success is True

            # list all items
            result = client.get("/tablespec")
            assert result.code == HTTP_OK
            assert result.success is True
            assert result.data["total"] == 1
            assert len(result.data["items"]) == 1

    def test_category(self, pokie_app):
        with pokie_app.test_client() as client:
            client = PokieClient(client)

            # list all items
            result = client.get("/catalog/category")
            assert result.code == HTTP_OK
            assert result.success is True
            assert result.data["total"] == 8
            assert len(result.data["items"]) == 8
            record = result.data["items"][0]

            # get first item
            result = client.get("/catalog/category/{}".format(record["id"]))
            assert result.code == HTTP_OK
            assert result.success is True

    def test_shipper(self, pokie_app):
        with pokie_app.test_client() as client:
            client = PokieClient(client)

            # list all items
            result = client.get("/catalog/shipper")
            assert result.code == HTTP_OK
            assert result.success is True
            assert result.data["total"] == 6
            assert len(result.data["items"]) == 6
            record = result.data["items"][0]

            # get first item
            result = client.get("/catalog/shipper/{}".format(record["id"]))
            assert result.code == HTTP_OK
            assert result.success is True

            # add new item
            record = {"id": 999, "name": "sample", "phone": "0000"}
            result = client.post("/catalog/shipper", data=record)
            assert result.code == HTTP_OK
            assert result.success is True

            # remove item
            result = client.delete("/catalog/shipper/{}".format(999))
            assert result.code == HTTP_OK
            assert result.success is True

    def test_states(self, pokie_app):
        with pokie_app.test_client() as client:
            client = PokieClient(client)

            # list all items
            result = client.get("/region/states")
            assert result.code == HTTP_OK
            assert result.success is True
            assert result.data["total"] == 51
            assert len(result.data["items"]) == 51
            record = result.data["items"][0]

            # get first item
            result = client.get("/region/states/{}".format(record["id"]))
            assert result.code == HTTP_OK
            assert result.success is True

            # insert item
            record = {"id": 128, "abbr": "NN", "name": "Narnia", "region": "closet"}
            result = client.post("/region/states", data=record)
            assert result.code == HTTP_OK
            assert result.success is True

            # remove item
            result = client.delete("/region/states/{}".format(128))
            assert result.code == HTTP_OK
            assert result.success is True

    def test_territories(self, pokie_app):
        with pokie_app.test_client() as client:
            client = PokieClient(client)

            # list all items
            result = client.get("/region/territories")
            assert result.code == HTTP_OK
            assert result.success is True

            assert result.data["total"] == 53
            assert len(result.data["items"]) == 53
            record = result.data["items"][0]

            # get first item
            result = client.get("/region/territories/{}".format(record["id"]))
            assert result.code == HTTP_OK
            assert result.success is True
            assert result.data["territory_description"] == "hamburger"

            # insert item
            id_record = "09999"
            record = {
                "id": id_record,
                "territory_description": "description",
                "region_id": 1,
            }
            result = client.post("/region/territories", data=record)
            assert result.code == HTTP_OK
            assert result.success is True

            # remove item
            result = client.delete("/region/territories/{}".format(id_record))
            assert result.code == HTTP_OK
            assert result.success is True

    def test_suppliers(self, pokie_app):
        with pokie_app.test_client() as client:
            client = PokieClient(client)

            # list all items
            result = client.get("/catalog/supplier")
            assert result.code == HTTP_OK
            assert result.success is True

            assert result.data["total"] == 29
            assert len(result.data["items"]) == 29
            record = result.data["items"][0]

            # get first item
            result = client.get("/catalog/supplier/{}".format(record["id"]))
            assert result.code == HTTP_OK
            assert result.success is True

            # insert item
            id_record = "09999"
            record = {
                "id": 9999,
                "company_name": "supplier 9999",
                "contact_name": "john connor",
                "homepage": "http://google.com",
            }

            result = client.post("/catalog/supplier", data=record)
            assert result.code == HTTP_OK
            assert result.success is True

            # remove item
            result = client.delete("/catalog/supplier/{}".format(id_record))
            assert result.code == HTTP_OK
            assert result.success is True



import pytest
from rick.serializer.json import ExtendedJsonEncoder

from pokie.constants import DI_DB
from pokie.http import DbGridRequest, JsonResponse, CamelCaseJsonResponse
from pokie_test.dto import CustomerRecord
from pokie_test.repository import CustomerRepository


@pytest.fixture
def result_dict():
    return {
        "field_a": "123",
        "field_b": "456",
    }


@pytest.fixture
def result_str():
    return "the quick brown fox jumps over the lazy dog"


@pytest.fixture
def result_list():
    return ["item_1", "item_2", "item_3"]


@pytest.fixture
def result_object(pokie_di):
    repo = CustomerRepository(pokie_di.get(DI_DB))
    return repo.fetch_pk("DUMON")


class TestResponse:
    def test_serializer(self):
        obj = JsonResponse()
        serializer = obj.serializer()()
        assert isinstance(serializer, ExtendedJsonEncoder)

    def test_response(self, result_dict):
        # success response
        obj = JsonResponse(result_dict)
        assert "error" not in obj.response.keys()
        assert result_dict == obj.response["data"]

        # error response, no specific error
        obj = JsonResponse(result_dict, success=False)
        assert "error" in obj.response.keys()
        assert isinstance(obj.response["error"], dict)
        assert obj.msg_default_error == obj.response["error"]["message"]

        # error response, override error
        # in this case, error has the payload of the parameter
        obj = JsonResponse(result_dict, success=False, error="some error")
        assert "error" in obj.response.keys()
        assert isinstance(obj.response["error"], str)
        assert "some error" == obj.response["error"]

    def test_dict(self, pokie_app, result_dict):
        obj = JsonResponse(result_dict)

        result = obj.assemble(pokie_app)
        assert (
            result.get_data(True)
            == '{"success":true,"data":{"field_a":"123","field_b":"456"}}'
        )

    def test_str(self, pokie_app, result_str):
        obj = JsonResponse(result_str)

        result = obj.assemble(pokie_app)
        assert (
            result.get_data(True)
            == '{"success":true,"data":"the quick brown fox jumps over the lazy dog"}'
        )

    def test_list(self, pokie_app, result_list):
        obj = JsonResponse(result_list)

        result = obj.assemble(pokie_app)
        assert (
            result.get_data(True)
            == '{"success":true,"data":["item_1","item_2","item_3"]}'
        )

    def test_object(self, pokie_app, result_object):
        obj = JsonResponse(result_object)

        result = obj.assemble(pokie_app)
        assert (
            result.get_data(True)
            == '{"success":true,"data":{"address":"67, rue des Cinquante Otages","city":"Nantes","company_name":"Du monde entier","contact_name":"Janine Labrune","contact_title":"Owner","country":"France","fax":"40.67.89.89","id":"DUMON","phone":"40.67.88.88","postal_code":"44000","region":null}}'
        )

    def test_compact(self, pokie_app, result_list):
        # test debug mode
        obj = JsonResponse(result_list)
        pokie_app.json.compact = True
        result = obj.assemble(pokie_app)
        assert (
            result.get_data(True)
            == '{\n  "success": true, \n  "data": [\n    "item_1", \n    "item_2", \n    "item_3"\n  ]\n}'
        )


class TestCamelCaseResponse:
    def serializer(self):
        obj = CamelCaseJsonResponse()
        serializer = obj.serializer()()
        assert isinstance(serializer, ExtendedJsonEncoder)

    def test_response(self, result_dict):
        # success response
        obj = CamelCaseJsonResponse(result_dict)
        assert "error" not in obj.response.keys()
        assert result_dict == obj.response["data"]

        # error response, no specific error
        obj = CamelCaseJsonResponse(result_dict, success=False)
        assert "error" in obj.response.keys()
        assert isinstance(obj.response["error"], dict)
        assert obj.msg_default_error == obj.response["error"]["message"]

        # error response, override error
        # in this case, error has the payload of the parameter
        obj = CamelCaseJsonResponse(result_dict, success=False, error="some error")
        assert "error" in obj.response.keys()
        assert isinstance(obj.response["error"], str)
        assert "some error" == obj.response["error"]

    def test_dict(self, pokie_app, result_dict):
        obj = CamelCaseJsonResponse(result_dict)

        result = obj.assemble(pokie_app)
        assert (
            result.get_data(True)
            == '{"success":true,"data":{"fieldA":"123","fieldB":"456"}}'
        )

    def test_str(self, pokie_app, result_str):
        obj = CamelCaseJsonResponse(result_str)

        result = obj.assemble(pokie_app)
        assert (
            result.get_data(True)
            == '{"success":true,"data":"the quick brown fox jumps over the lazy dog"}'
        )

    def test_list(self, pokie_app, result_list):
        obj = CamelCaseJsonResponse(result_list)

        result = obj.assemble(pokie_app)
        assert (
            result.get_data(True)
            == '{"success":true,"data":["item_1","item_2","item_3"]}'
        )

    def test_object(self, pokie_app, result_object):
        obj = CamelCaseJsonResponse(result_object)

        result = obj.assemble(pokie_app)
        assert (
            result.get_data(True)
            == '{"success":true,"data":{"address":"67, rue des Cinquante Otages","city":"Nantes","companyName":"Du monde entier","contactName":"Janine Labrune","contactTitle":"Owner","country":"France","fax":"40.67.89.89","id":"DUMON","phone":"40.67.88.88","postalCode":"44000","region":null}}'
        )

    def test_compact(self, pokie_app, result_list):
        # test debug mode
        obj = CamelCaseJsonResponse(result_list)
        pokie_app.json.compact = True
        result = obj.assemble(pokie_app)
        assert (
            result.get_data(True)
            == '{\n  "success": true, \n  "data": [\n    "item_1", \n    "item_2", \n    "item_3"\n  ]\n}'
        )



from .view import PokieView, PokieAuthView
from .routes import AutoRouter
from .http_error import HttpErrorHandler
from .response import ResponseRendererInterface, JsonResponse, CamelCaseJsonResponse
from .dbgrid import DbGridRequest



from rick_db import DbGrid, Repository


class RestServiceMixin:
    def get(self, id_record):
        return self.repository.fetch_pk(id_record)

    def delete(self, id_record):
        return self.repository.delete_pk(id_record)

    def insert(self, record):
        return self.repository.insert_pk(record)

    def update(self, id_record, record):
        return self.repository.update(record, id_record)

    def exists(self, id_record):
        return self.repository.valid_pk(id_record)

    def list(
        self,
        search_fields: list = None,
        search_text: str = None,
        match_fields: dict = None,
        limit: int = None,
        offset: int = None,
        sort_fields: dict = None,
        search_filter: list = None,
    ):
        grid = DbGrid(self.repository, search_fields, DbGrid.SEARCH_ANY)
        return grid.run(
            None,
            search_text=search_text,
            match_fields=match_fields,
            limit=limit,
            offset=offset,
            sort_fields=sort_fields,
            search_fields=search_filter,
        )

    @property
    def repository(self) -> Repository:
        raise RuntimeError("RestServiceMixin::repository must be overridden")



from rick.form import RequestRecord, field
from pokie.http import PokieView


class SampleRequestRecord(RequestRecord):
    fields = {
        "field1": field(validators="required|numeric|maxlen:3"),
        "field2": field(validators="required|maxlen:3"),
    }


class CustomRequestRecordView(PokieView):
    # RequestRecord class for body operations
    # the view will automatically use this class to process the body and - if successful - make the results available
    # at self.request
    request_class = SampleRequestRecord

    def get(self):
        # just used for testing purposes
        return self.success()

    def post(self):
        # if SampleRequestRecord validation is successful, the object is available in self.request;
        # if validation is not successful, this method isn't even executed
        #
        # generate a result based on the self.request RequestRecord object
        result = {
            "field1_contents": self.request.get("field1"),
            "field2_contents": self.request.get("field2"),
        }
        return self.success(result)

    def put(self):
        # if SampleRequestRecord validation is successful, the object is available in self.request;
        # if validation is not successful, this method isn't even executed
        #
        # generate a result based on the self.request RequestRecord object
        result = {
            "field1_contents": self.request.get("field1"),
            "field2_contents": self.request.get("field2"),
        }
        return self.success(result)

    def patch(self):
        # if SampleRequestRecord validation is successful, the object is available in self.request;
        # if validation is not successful, this method isn't even executed
        #
        # generate a result based on the self.request RequestRecord object
        result = {
            "field1_contents": self.request.get("field1"),
            "field2_contents": self.request.get("field2"),
        }
        return self.success(result)



import json
from typing import Type
from collections.abc import Mapping
import humps
from rick.serializer.json.json import CamelCaseJsonEncoder, ExtendedJsonEncoder
from pokie.constants import HTTP_OK


class ResponseRendererInterface:
    def __init__(
        self,
        data: dict = None,
        success: bool = True,
        error: dict = None,
        code: int = HTTP_OK,
        mime_type: str = None,
        headers: list = None,
    ):
        pass

    def assemble(self, _app, **kwargs):
        pass


class JsonResponse(ResponseRendererInterface):
    """
    Default JSON response formatter

    The usual JSON response has the following format:

    success is True:
    {
        "success": True,
        "data": {...}
    }

    success is False:
    {
        "success": False,
        "error": {
            "message": "...",
            [...optional extra keys, such as formError...]
        }
    }
    """

    # mime type
    mime_type = "application/json"
    # default error message
    msg_default_error = "an error has occurred"

    def __init__(
        self,
        data: dict = None,
        success: bool = True,
        error: dict = None,
        code: int = HTTP_OK,
        mime_type: str = None,
        headers: list = None,
    ):
        """
        Constructor for standardized json response
        :param data:
        :param success:
        :param error:
        :param code:
        :param mime_type:
        :param headers:
        """
        self.headers = headers
        self.code = code
        self.response = {"success": success}

        # override default mime type
        if mime_type is not None:
            self.mime_type = mime_type

        # success always has 'data' object
        if success and data is None:
            data = {}

        # error always has 'error' object
        if not success and error is None:
            error = {"message": self.msg_default_error}
        else:
            if isinstance(error, Mapping):
                if "message" not in error.keys():
                    error["message"] = self.msg_default_error

        if data is not None:
            self.response["data"] = data

        if error is not None:
            self.response["error"] = error

    def assemble(self, _app, **kwargs):
        """
        Assemble Flask response object
        :param _app:
        :return: Response
        """
        indent = None
        separators = (",", ":")

        if _app.json.compact or _app.debug:
            indent = 2
            separators = (", ", ": ")

        data = json.dumps(
            self.response, indent=indent, separators=separators, cls=self.serializer()
        )
        return _app.response_class(
            data, status=self.code, mimetype=self.mime_type, headers=self.headers
        )

    def serializer(self) -> Type[json.JSONEncoder]:
        """
        Get JSON serializer
        :return:
        """
        return ExtendedJsonEncoder


class CamelCaseJsonResponse(JsonResponse):
    def assemble(self, _app, **kwargs):
        """
        Assemble Flask response object
        :param _app:
        :return: Response
        """
        indent = None
        separators = (",", ":")

        if _app.json.compact or _app.debug:
            indent = 2
            separators = (", ", ": ")

        data = json.dumps(
            humps.camelize(self.response),
            indent=indent,
            separators=separators,
            cls=self.serializer(),
        )
        return _app.response_class(
            data, status=self.code, mimetype=self.mime_type, headers=self.headers
        )

    def serializer(self) -> Type[json.JSONEncoder]:
        """
        Get JSON serializer
        :return:
        """
        return CamelCaseJsonEncoder



import os
import sys
import threading
from argparse import ArgumentParser
from typing import List

from flask import Flask
from rick.base import Di, Container, MapLoader
from rick.event import EventManager
from rick.mixin import Injectable
from rick.util.loader import load_class
from rick.resource.console import ConsoleWriter

from pokie.constants import (
    DI_CONFIG,
    DI_SERVICES,
    DI_FLASK,
    DI_APP,
    DI_EVENTS,
    DI_TTY,
    DI_SIGNAL,
    CFG_HTTP_ERROR_HANDLER,
    DI_HTTP_ERROR_HANDLER,
)
from .middleware import ModuleRunnerMiddleware
from .module import BaseModule
from .command import CliCommand
from pokie.util.cli_args import ArgParser
from .signal import SignalManager


class FlaskApplication:
    CLI_CMD_SUCCESS = 0
    CLI_CMD_FAILED = 1
    CLI_CMD_NOT_FOUND = 2

    module_file_name = "module"  # module class file name
    module_class_name = "Module"  # default module class name

    system_modules = [
        "pokie.contrib.base",
    ]  # system modules to always be included

    def __init__(self, cfg: Container):
        self.di = Di()
        self.app = None
        self.modules = {}  # app module list

        self.di.add(DI_CONFIG, cfg)
        self.di.add(DI_APP, self)
        self.cfg = cfg
        self.lock = threading.Lock()
        self.initialized = False

        self.pre_http_hooks = (
            []
        )  # list of hooks to run before initializing http operations
        self.pre_cli_hooks = (
            []
        )  # list of hooks to run before initializing cli operations

    def build(self, module_list: list, factories: List = None) -> Flask:
        """
        Build the application

        Factories is a list of optional callables to assemble functionality on top of Di, eg. database connection,
        cache, logging, etc. Factories are called *before* modules are initialized, to ensure all required dependencies
        are available

        :param module_list: list of module names to initialize
        :param factories: optional list of callables to be initialized with the application
        :return:
        """
        if not factories:
            factories = []

        self.app = Flask(type(self).__name__, static_folder=None)

        self.app.di = self.di
        self.di.add(DI_FLASK, self.app)

        # initialize signal manager
        self.di.add(DI_SIGNAL, SignalManager(self.di))

        # initialize TTY
        self.di.add(DI_TTY, ConsoleWriter())

        # run factories
        for factory in factories:
            if type(factory) is str:
                # if factory is string, assume it is a path to a callable
                factory = load_class(factory, raise_exception=True)
            if not callable(factory):
                raise RuntimeError("build(): non-callable or non-existing factory")
            else:
                factory(self.di)

        # load modules
        self.modules = {}
        module_list = [*self.system_modules, *module_list]
        for name in module_list:
            cls = load_class(
                "{}.{}.{}".format(name, self.module_file_name, self.module_class_name),
                raise_exception=True,
            )
            if cls is None:
                raise RuntimeError(
                    "build(): cannot load module '{}' - Module() class not found".format(
                        name
                    )
                )
            if not issubclass(cls, BaseModule):
                raise RuntimeError(
                    "build(): Class Module on '{}' must extend BaseModule".format(name)
                )
            if name in self.modules.keys():
                raise ValueError(
                    "build(): Module named '{}' already exists".format(name)
                )
            self.modules[name] = cls(self.di)

        # build service map
        svc_map = {}
        for name, m in self.modules.items():
            services = getattr(m, "services", {})
            if type(services) is dict:
                svc_map.update(services)
            else:
                raise RuntimeError(
                    "build(): cannot load service map from module '{}'; attribute must be of type dict".format(
                        name
                    )
                )
        # register service mapper
        self.di.add(DI_SERVICES, MapLoader(self.di, svc_map))

        # parse events from modules
        evt_mgr = EventManager()
        for _, module in self.modules.items():
            module_events = getattr(module, "events", None)
            if isinstance(module_events, dict):
                for evt_name, evt_details in module_events.items():
                    for priority, handlers in evt_details.items():
                        for handler in handlers:
                            evt_mgr.add_handler(evt_name, handler, int(priority))

        self.di.add(DI_EVENTS, evt_mgr)

        # register exception handler
        if self.cfg.has(CFG_HTTP_ERROR_HANDLER):
            handler = load_class(
                self.cfg.get(CFG_HTTP_ERROR_HANDLER), raise_exception=True
            )
            if not issubclass(handler, Injectable):
                raise RuntimeError(
                    "build(): HTTP_ERROR_HANDLER class does not extend Injectable"
                )
            # initialize & register handler
            handler = handler(self.di)
            self.di.add(DI_HTTP_ERROR_HANDLER, handler)

        self.app.wsgi_app = ModuleRunnerMiddleware(self.app.wsgi_app, self)
        return self.app

    def register_pre_http_hook(self, f):
        """
        Register a hook to be executed during the init() of the webserver

        the hook must have the following interface:

        callable(app:FlaskApplication)

        :param f:
        :return:
        """
        self.pre_http_hooks.append(f)

    def register_pre_cli_hook(self, f):
        """
        Register a hook to be executed before any cli operation

        the hook must have the following interface:

        callable(app:FlaskApplication)

        :param f:
        :return:
        """
        self.pre_cli_hooks.append(f)

    def init(self):
        with self.lock:
            if not self.initialized:
                # initialize modules
                for _, module in self.modules.items():
                    module.build(self)
                self.initialized = True

                # call pre-http hooks
                for fn in self.pre_http_hooks:
                    fn(self)

            def stub(**kwargs):
                pass

            # instead of flag, we empty the method
            setattr(self, "init", stub)

    def http(self, **kwargs):
        self.app.run(**kwargs)

    def cli_runner(self, command: str, args: list = None, **kwargs) -> int:
        # run pre-cli hooks
        for fn in self.pre_cli_hooks:
            fn(self)

        # either console or inline commands
        if args is None:
            args = []

        # parameter parser
        parser = ArgParser(**kwargs)

        if "writer" in kwargs.keys():
            tty = kwargs["writer"]
        else:
            tty = ConsoleWriter()

        # lookup handler
        for _, module in self.modules.items():
            if command in module.cmd.keys():
                # load_class may raise ModuleNotFoundError if path not found
                handler = load_class(module.cmd[command], raise_exception=True)

                if not handler:
                    raise RuntimeError(
                        "cli(): handler class '{}' not found".format(
                            module.cmd[command]
                        )
                    )

                if not issubclass(handler, CliCommand):
                    raise RuntimeError(
                        "cli(): command handler does not extend CliCommand"
                    )

                handler = handler(self.di, writer=tty)  # type: CliCommand
                if not handler.skipargs:  # skipargs controls usage of argparser
                    handler.arguments(parser)
                    args = parser.parse_args(args)
                    if parser.failed:
                        # invalid/insufficient args
                        tty.error(parser.error_message)
                        parser.print_help(tty.stderr)
                        return self.CLI_CMD_FAILED
                else:
                    # skipargs is true, all argparsing is ignored
                    # this allow for custom cli arg handling
                    args = None

                if handler.run(args):
                    return self.CLI_CMD_SUCCESS
                return self.CLI_CMD_FAILED

        # command not found
        tty.error("error executing '{}': command not found".format(command))
        return self.CLI_CMD_NOT_FOUND

    def cli(self, **kwargs):
        """
        Execute CLI commands
        :param kwargs: optional parameters for ArgumentParse
        :return:
        """
        # default command when no args detected
        command = "list"
        # extract command if specified
        if len(sys.argv) > 1:
            command = str(sys.argv[1])

        if "add_help" not in kwargs.keys():
            kwargs["add_help"] = False
        if "usage" not in kwargs.keys():
            kwargs["usage"] = "{} {} [OPTIONS...]".format(
                os.path.basename(sys.argv[0]), command
            )

        # exit code directly maps return codes
        exit(self.cli_runner(command, sys.argv[2:], **kwargs))



from typing import Optional

from rick.mixin.injectable import Injectable
from rick.util.loader import load_class

from pokie.contrib.auth.service.user import AuthUser
from pokie.constants import DI_CONFIG
from pokie.plugins.auth import AuthPluginInterface


class AuthService(Injectable):
    def authenticate(
        self, username: str, password: str, **kwargs
    ) -> Optional[AuthUser]:
        for plugin in self.auth_plugins:  # type: AuthPluginInterface
            if plugin.valid_username(username, **kwargs):
                result = plugin.autenticate(username, password, **kwargs)
                if result is not None:
                    return result
        return None

    def update_password(self, username: str, password: str, **kwargs) -> bool:
        for plugin in self.auth_plugins:  # type: AuthPluginInterface
            if plugin.valid_username(username) and plugin.has_capability(
                AuthPluginInterface.UPDATE_PASSWORD
            ):
                return plugin.update_password(username, password, **kwargs)
        return False

    def load_id(self, id_user, plugin_cls=None, **kwargs) -> Optional[AuthUser]:
        """
        Attempts to find a user profile by id, using the registed plugins
        :param id_user: unique user identifier
        :param plugin_cls: optional class of plugin to use
        :return: AuthUser object or None
        """
        for plugin in self.auth_plugins:  # type: AuthPluginInterface
            if plugin_cls is None or isinstance(plugin, plugin_cls):
                result = plugin.load_id(id_user, **kwargs)
                if result is not None:
                    return result
        return None

    @property
    def auth_plugins(self) -> list:
        di = self.get_di()
        cfg = di.get(DI_CONFIG)
        plugins = []
        auth_plugins = cfg.get("auth_plugins", [])
        if len(auth_plugins) == 0:
            raise RuntimeError(
                "AuthService: authentication plugins missing from configuration"
            )
        for name in auth_plugins:
            plugin = load_class(name)
            if plugin is None:
                raise RuntimeError(
                    "AuthService: auth plugin '{}' not found".format(name)
                )
            if not issubclass(plugin, AuthPluginInterface):
                raise RuntimeError(
                    "AuthService: auth plugin '{}' must implement AuthPlugin interface".format(
                        name
                    )
                )
            plugins.append(plugin(di))
        return plugins



from pokie.rest import RestView


class TerritoriesView(RestView):
    def get(self, id_record=None):
        if id_record is None:
            return self.list()

        record = self.svc.get(id_record)
        if record is None:
            return self.not_found()

        # this just returns hamburger in description
        record.territory_description = "hamburger"
        return self.success(record)



from pokie.contrib.auth.constants import SVC_USER, SVC_ACL, SVC_AUTH
from pokie.core import BaseModule


class Module(BaseModule):
    name = "auth"
    description = "Authentication module"

    services = {
        SVC_AUTH: "pokie.contrib.auth.service.AuthService",
        SVC_ACL: "pokie.contrib.auth.service.AclService",
        SVC_USER: "pokie.contrib.auth.service.UserService",
    }

    cmd = {
        # user-related operations
        "user:create": "pokie.contrib.auth.cli.UserCreateCmd",
        "user:info": "pokie.contrib.auth.cli.UserInfoCmd",
        "user:mod": "pokie.contrib.auth.cli.UserModCmd",
        "user:list": "pokie.contrib.auth.cli.UserListCmd",
        "user:role": "pokie.contrib.auth.cli.AclUserRoleCmd",
        # acl operations
        "role:list": "pokie.contrib.auth.cli.AclRoleListCmd",
        "role:create": "pokie.contrib.auth.cli.AclRoleCreateCmd",
        "role:remove": "pokie.contrib.auth.cli.AclRoleRemoveCmd",
        "role:info": "pokie.contrib.auth.cli.AclRoleInfoCmd",
        "role:link": "pokie.contrib.auth.cli.AclRoleLinkCmd",
        "role:unlink": "pokie.contrib.auth.cli.AclRoleUnlinkCmd",
        "resource:list": "pokie.contrib.auth.cli.AclResourceListCmd",
        "resource:create": "pokie.contrib.auth.cli.AclResourceCreateCmd",
        "resource:link": "pokie.contrib.auth.cli.AclResourceLinkCmd",
        "resource:unlink": "pokie.contrib.auth.cli.AclResourceUnlinkCmd",
    }

    def build(self, parent=None):
        pass



from rick_db.conn import Connection
from rick_db.repository import BaseRepository
from rick_db.sql import Select


class ValidatorRepository(BaseRepository):
    def __init__(self, db: Connection):
        """
        Constructor
        This class is a stub repository; it doesn't have Record class, table name or schema
        :param db:
        """
        super(ValidatorRepository, self).__init__(db, "")

    def pk_exists(self, pk_value, pk_name, table_name, schema=None) -> bool:
        sql, values = (
            Select(self._dialect)
            .from_(table_name, cols=[pk_name], schema=schema)
            .where(pk_name, "=", pk_value)
            .assemble()
        )
        with self._db.cursor() as c:
            record = c.fetchone(sql, values)
            return record is not None



import sys
import pytest
import os
from rick_db.conn.pg import PgConnection
from .client import PokieClient
from pokie.constants import (
    POKIE_NAMESPACE,
    POKIE_APP,
    POKIE_FACTORY,
    DI_CONFIG,
    DI_DB,
    DI_SERVICES,
    CFG_TEST_DB_REUSE,
    CFG_TEST_SKIP_MIGRATIONS,
    CFG_TEST_SKIP_FIXTURES,
    CFG_TEST_DB_NAME,
    CFG_TEST_MANAGE_DB,
    CFG_TEST_SHARE_CTX,
    CFG_TEST_DB_HOST,
    CFG_TEST_DB_PORT,
    CFG_TEST_DB_USER,
    CFG_TEST_DB_PASSWORD,
    CFG_TEST_DB_SSL,
    DI_APP,
)


@pytest.hookimpl()
def pytest_addoption(parser) -> None:
    group = parser.getgroup("pokie")
    group.addoption(
        "--reuse-db",
        action="store_true",
        dest="reuse_db",
        default=False,
        help="Do not re-create test database between tests, and re-use database if already exists.",
    )
    group.addoption(
        "--no_migrations",
        action="store_true",
        dest="skip_migrations",
        default=False,
        help="Do not run database migrations",
    )
    group.addoption(
        "--no_fixtures",
        action="store_true",
        dest="skip_fixtures",
        default=False,
        help="Do not run fixtures",
    )


@pytest.fixture(scope="session")
def pokie_factory():
    """
    Base pokie initialization

    attempts to extract the Flask 'app' object and the optional factory 'build_pokie()' from the main scope of the
    application
    """
    # attempt to extract Flask application from the main namespace
    pokie_ns = os.getenv(POKIE_NAMESPACE, "__main__")
    pokie_app = os.getenv(POKIE_APP, "app")
    pokie_factory = os.getenv(POKIE_FACTORY, "build_pokie")

    # validate namespace
    if pokie_ns not in sys.modules.keys():
        raise RuntimeError("Error: cannot find pokie namespace '{}'".format(pokie_ns))

    app = getattr(sys.modules[pokie_ns], pokie_app)
    # validate object
    if app is None:
        raise RuntimeError(
            "Error: cannot find flask object '{}' in namespace '{}'".format(
                pokie_app, pokie_ns
            )
        )

    # validate factory if exists
    factory = getattr(sys.modules[pokie_ns], pokie_factory)
    if factory is not None:
        if not callable(factory):
            raise RuntimeError("Error: attribute named 'build_pokie' is not callable")

    return app, factory


@pytest.fixture()
def pokie_config(pokie_app):
    """
    Application config fixture
    """
    yield pokie_app.di.get(DI_CONFIG)


@pytest.fixture()
def pokie_di(pokie_app):
    """
    Application Di fixture
    """
    yield pokie_app.di


@pytest.fixture()
def pokie_db(pokie_app):
    """
    Database Connection fixture
    """
    yield pokie_app.di.get(DI_DB)


@pytest.fixture()
def pokie_service_manager(pokie_app):
    """
    Service Manager fixture
    """
    yield pokie_app.di.get(DI_SERVICES)


@pytest.fixture()
def pokie_client(pokie_app):
    """
    Barebones REST client
    """
    with pokie_app.test_client() as client:
        yield PokieClient(client)


@pytest.fixture(autouse=True)
def pokie_app(request, pokie_factory):
    """
    App initialization fixture

    Pokie config parameters are extracted from TestConfigTemplate
    """
    app, factory = pokie_factory
    cfg = app.di.get(DI_CONFIG)

    # pytest arguments can override the base setup
    if request.config.getvalue("reuse_db"):
        reuse_db = True
    else:
        reuse_db = cfg.get(CFG_TEST_DB_REUSE)

    if request.config.getvalue("skip_migrations"):
        skip_migrations = True
    else:
        skip_migrations = cfg.get(CFG_TEST_SKIP_MIGRATIONS)

    if request.config.getvalue("skip_fixtures"):
        skip_fixtures = True
    else:
        skip_fixtures = cfg.get(CFG_TEST_SKIP_FIXTURES)

    # -- app context
    if not cfg.get(CFG_TEST_SHARE_CTX):
        # if context isn't shared, create whole new application
        _, app = factory()

    # if db management is disabled, skip db management altogether
    if not cfg.get(CFG_TEST_MANAGE_DB):
        yield app
        return

    # -- db stuff
    test_db = cfg.get(CFG_TEST_DB_NAME, "test_pokie")
    try:
        # DI_DB container is replaced automatically
        _init_test_db(app, test_db, reuse_db, skip_migrations, skip_fixtures)

    except Exception as e:
        raise RuntimeError(
            "Error initializing database; does the test user has database create/drop privileges? Error: {}".format(
                str(e)
            )
        )

    yield app

    # cleanup
    if not reuse_db:
        conn = app.di.get(DI_DB)
        if conn:
            # redo connection for drop purposes
            conn = _test_db_connection(app)
            conn.metadata().drop_database(test_db)


def _test_db_connection(app, db_name: str = None):
    cfg = app.di.get(DI_CONFIG)
    db_cfg = {
        "dbname": "postgres" if db_name is None else db_name,
        "host": cfg.get(CFG_TEST_DB_HOST, "localhost"),
        "port": int(cfg.get(CFG_TEST_DB_PORT, 5432)),
        "user": cfg.get(CFG_TEST_DB_USER, "postgres"),
        "password": cfg.get(CFG_TEST_DB_PASSWORD, ""),
        "sslmode": None if not cfg.get(CFG_TEST_DB_SSL, "1") else "require",
    }
    return PgConnection(**db_cfg)


def _init_test_db(app, test_db, reuse_db, skip_migrations, skip_fixtures):
    if not test_db:
        # if test database name is empty, skip database handling altogether
        app.di.add(DI_DB, None)
        return None

    # first connection for administrative purposes
    conn = _test_db_connection(app, None)
    mgr = conn.metadata()
    db_exists = mgr.database_exists(test_db)
    if db_exists and not reuse_db:
        mgr.drop_database(test_db)
        db_exists = False

    if not db_exists:
        mgr.create_database(test_db)

    # actual final connection
    conn = _test_db_connection(app, test_db)
    app.di.add(DI_DB, conn, replace=True)

    if not db_exists:
        pokie_main = app.di.get(DI_APP)
        pokie_main.cli_runner("db:init")

        if not skip_migrations:
            pokie_main.cli_runner("db:update")

        if not skip_fixtures:
            pokie_main.cli_runner("fixture:run")



from rick.base import Di
from rick.resource.redis import RedisCache

from pokie.constants import (
    CFG_REDIS_HOST,
    DI_CONFIG,
    CFG_REDIS_PORT,
    DI_REDIS,
    CFG_REDIS_PASSWORD,
    CFG_REDIS_DB,
    CFG_REDIS_SSL,
)


def RedisFactory(_di: Di):
    """
    REDIS factory
    Note: The connection is only created when the resource is accessed on Di
    :param _di:
    :return:
    """

    @_di.register(DI_REDIS)
    def _factory(_di: Di):
        cfg = _di.get(DI_CONFIG)
        redis_cfg = {
            "host": cfg.get(CFG_REDIS_HOST, "localhost"),
            "port": int(cfg.get(CFG_REDIS_PORT, 6379)),
            "password": cfg.get(CFG_REDIS_PASSWORD, ""),
            "db": int(cfg.get(CFG_REDIS_DB, 0)),
            "ssl": True if cfg.get(CFG_REDIS_SSL, None) else False,
        }
        return RedisCache(**redis_cfg)



import secrets
import token
from typing import List, Optional

from rick.base import Di
from rick.form import RequestRecord
from rick_db import fieldmapper
from rick_db.mapper import ATTR_RECORD_MAGIC, ATTR_TABLE, ATTR_SCHEMA, BaseRecord
from rick_db.util.pg import PgInfo

import pokie.codegen.pg
from pokie.codegen import RequestGenerator
from pokie.codegen.pg import PgTableSpec
from pokie.constants import DI_DB
from pokie.http import PokieView, AutoRouter

from pokie.rest import RestView


class Auto:
    @staticmethod
    def rest(
        app: object,
        slug: str,
        dto_record: object,
        request_class: RequestRecord = None,
        service: str = None,
        id_type: str = None,
        search_fields: list = None,
        allow_methods: list = None,
        base_cls: tuple = None,
        mixins: tuple = None,
        **kwargs
    ):
        """
        Assemble and register a RESTView for the specified DTO Record

        A View class is generated programmatically, based on RestView (by default). The View class may inherit other
        mixins, e.g. to provide authentication or custom business logic.
        The generated View class is registered with the application, using the slug as endpoint.


        - If service is specified, it is used instead of the automatically generated service;
        - For search to work on listing operations, search_fields must contain the field names;
        - The base View class is RestView, but it can be overridden by the base_cls parameter;
        - If other base class is used, it must extend from PokieView;

        :param app: Flask object
        :param dto_record: Record to use
        :param request_class: optional Request class
        :param service: optional service name
        :param id_type: optional id type for the parameter (to be passed to Flask, e.g. "int"-> id:<int:id_value>)
        :param search_fields: optional list of field names to perform search
        :param allow_methods: optional list of http methods to allow
        :param base_cls: optional base class to use instead of RestView
        :param mixins: optional list of mixins to include
        :param kwargs: optional extra parameters
        :return:
        """

        view = Auto._view_factory(
            app.di,
            dto_record,
            request_class,
            service,
            search_fields,
            allow_methods,
            base_cls,
            mixins,
            **kwargs
        )
        AutoRouter.resource(app, slug, view, id_type=id_type)
        return view

    @staticmethod
    def view(
        app: object,
        table_name: str,
        schema: str = None,
        search_fields: List = None,
        camel_case: bool = False,
        allow_methods: list = None,
        base_cls: tuple = None,
        mixins: tuple = None,
        **kwargs
    ) -> PokieView:
        if not schema:
            schema = PgInfo.SCHEMA_DEFAULT

        db = app.di.get(DI_DB)
        info = PgInfo(db)
        if not info.table_exists(table_name, schema=schema):
            raise ValueError(
                "Auto.view(): table name '{}' not found in schema '{}'".format(
                    table_name, schema
                )
            )

        spec = PgTableSpec(db).generate(table_name, schema)
        if search_fields is None:
            search_fields = [
                f.name for f in spec.fields if f.dtype in ["varchar", "text"]
            ]

        request_class = pokie.codegen.RequestGenerator().generate_class(
            spec, camelcase=camel_case
        )
        record_class = pokie.codegen.RecordGenerator().generate_class(
            spec, camelcase=camel_case
        )

        if not base_cls:
            base_cls = (RestView,)

        if not isinstance(
            base_cls,
            (
                tuple,
                list,
            ),
        ):
            base_cls = (base_cls,)

        if not mixins:
            mixins = ()
        extends = base_cls + mixins
        cls_attrs = {
            "request_class": request_class,
            "record_class": record_class,
            "search_fields": search_fields,
        }
        if allow_methods:
            cls_attrs["allow_methods"] = allow_methods

        cls = type("AutoView_{}".format(secrets.token_hex(8)), extends, cls_attrs)
        return Auto._patch_view_class(cls, mixins)

    @staticmethod
    def _view_factory(
        di: Di,
        dto_record: BaseRecord,
        request_class: RequestRecord,
        service: str = None,
        search_fields: list = None,
        allow_methods: list = None,
        base_cls: tuple = None,
        mixins: tuple = None,
        **kwargs
    ):
        if not base_cls:
            base_cls = (RestView,)
        if not mixins:
            mixins = ()
        if not isinstance(
            base_cls,
            (
                tuple,
                list,
            ),
        ):
            base_cls = (base_cls,)
        extends = base_cls + mixins
        cls_attrs = {
            "record_class": dto_record,
            "search_fields": search_fields,
        }
        if not request_class:
            # no request_class provided, attempt to build one from dto_record
            request_class = Auto._assemble_request_class(di, dto_record)

        if not request_class:
            raise ValueError(
                "Auto._view_factory(): could not assemble a request_class class; a explicit class is required"
            )
        cls_attrs["request_class"] = request_class

        # disable automatic service creation, use custom one instead
        if service:
            cls_attrs["service_name"] = service

        # only override if value specified
        if allow_methods:
            cls_attrs["allow_methods"] = allow_methods

        cls_attrs = {**cls_attrs, **kwargs}
        view_class = type(
            "AutoRestClass_{}".format(secrets.token_hex(8)), extends, cls_attrs
        )
        if not issubclass(view_class, PokieView):
            raise ValueError(
                "Automatic REST classing must use a base class inherited from RestView"
            )
        return Auto._patch_view_class(view_class, mixins)

    @staticmethod
    def _assemble_request_class(
        di: Di, dto_record: BaseRecord
    ) -> Optional[RequestRecord]:
        if getattr(dto_record, ATTR_RECORD_MAGIC) is True:
            table = getattr(dto_record, ATTR_TABLE)
            schema = getattr(dto_record, ATTR_SCHEMA)

            # found a table name, lets assume it is actually a db table
            if table:
                pg_spec = PgTableSpec(di.get(DI_DB))
                spec = pg_spec.generate(table, schema)
                return RequestGenerator().generate_class(spec)
        return None

    @staticmethod
    def _patch_view_class(view_class, mixins):
        """
        Compose mixin dispatch hooks, internal hooks and init_methods to final class
        :param view_class:
        :param mixins:
        :return:
        """
        for item in mixins:
            if getattr(item, "dispatch_hooks", None):
                view_class.dispatch_hooks.extend(item.dispatch_hooks)
            if getattr(item, "internal_hooks", None):
                view_class.internal_hooks.extend(item.internal_hooks)
            if getattr(item, "init_methods", None):
                view_class.init_methods.extend(item.init_methods)
        return view_class



from rick.base import Di

from pokie.constants import DI_DB
from rick.mixin import Injectable
from rick_db import Repository
from .service_mixin import RestServiceMixin


class RestService(Injectable, RestServiceMixin):
    record_class = None  # record class
    repository_class = None  # optional custom repository class

    def __init__(self, di: Di):
        super().__init__(di)
        # copy class-attributes to instance-attributes
        # this allows usage of multiple instances of RestService with different record and repository classes
        self._record_cls = self.record_class
        self._repository_cls = self.repository_class

    def set_record_class(self, cls):
        self._record_cls = cls

    def set_repository_class(self, cls):
        self._repository_cls = cls

    @property
    def repository(self) -> Repository:
        if self._record_cls is None:
            raise RuntimeError("Missing record class for repository")
        if self._repository_cls is None:
            return Repository(self.get_di().get(DI_DB), self._record_cls)
        else:
            return self._repository_cls(self.get_di().get(DI_DB))



from rick.base import Di
from rick_db.conn.pg import PgConnection

from pokie.constants import (
    CFG_DB_NAME,
    CFG_DB_HOST,
    CFG_DB_PORT,
    CFG_DB_USER,
    CFG_DB_PASSWORD,
    CFG_DB_SSL,
    DI_DB,
    DI_CONFIG,
)


def PgSqlFactory(_di: Di):
    """
    PostgreSQL connection factory
    Note: The connection is only created when the resource is accessed on Di
    """

    @_di.register(DI_DB)
    def _factory(_di: Di):
        cfg = _di.get(DI_CONFIG)
        db_cfg = {
            "dbname": cfg.get(CFG_DB_NAME, "postgres"),
            "host": cfg.get(CFG_DB_HOST, "localhost"),
            "port": int(cfg.get(CFG_DB_PORT, 5432)),
            "user": cfg.get(CFG_DB_USER, "postgres"),
            "password": cfg.get(CFG_DB_PASSWORD, ""),
            "sslmode": None if not cfg.get(CFG_DB_SSL, "1") else "require",
        }
        return PgConnection(**db_cfg)



import importlib
import os
import sys
from pokie.core import CliCommand


class PyTestCmd(CliCommand):
    description = "run pytest"
    skipargs = True

    def run(self, args) -> bool:
        if importlib.util.find_spec("pytest") is None:
            self.tty.error(
                "Pytest package not found; to use this command please install pytest"
            )
            return False

        args = []
        if len(sys.argv) > 2:
            args = sys.argv[2:]
        self.tty.write(
            self.tty.colorizer.white("[Pokie]", attr="bold")
            + " Running pytest with: {}".format(str(args))
        )
        import pytest

        sys.exit(pytest.main(args, plugins=["pokie.test"]))
